#include <stdio.h>
#include "Pixelbit.h"

CPixelbit::CPixelbit(){
	m_PixelAdr = NULL;
	m_ColorKey = 0xffffffff;
	m_Width = 0;
	m_Height = 0;
	Clear(1, 1);
}

CPixelbit::CPixelbit(int w, int h){
	m_PixelAdr = NULL;
	m_ColorKey = 0xffffffff;
	m_Width = 0;
	m_Height = 0;
	Clear(w, h);
}

CPixelbit::~CPixelbit(){
	if(m_PixelAdr != NULL){
		SelectObject(m_BmpHdc, m_BmpHndOld);
		DeleteObject(m_BmpHnd);
		DeleteDC(m_BmpHdc);
		m_PixelAdr = NULL;
	}
}

bool CPixelbit::Assign(void *ptr){
	int w, h;
	LPBITMAPINFOHEADER head;
	DWORD ofset;
	BYTE *adr;
	if(!ptr) return false;
	head = (LPBITMAPINFOHEADER)((BYTE *)ptr+sizeof(BITMAPFILEHEADER));
	if(head->biBitCount != 24 && head->biBitCount != 32) return false;
	CopyMemory(&ofset, (BYTE *)ptr+10, 4);
	w = head->biWidth;
	h = head->biHeight;
	if(w*h<1) return false;
	if(!Clear(w, h)) return false;
	adr = (BYTE *)ptr+ofset;
	switch(head->biBitCount){
	case 24: Set4BGR24(0, 0, m_Width, m_Height, adr); break;
	case 32: Set4BGR32(0, 0, m_Width, m_Height, adr); break;
	}
	return true;
}

bool CPixelbit::Clear(int w, int h){
	DWORD *adr = NULL;
	BITMAPINFO bmpinfo;
	HBITMAP bmphnd;
	HDC hdc, hdcnew;
	if(w*h<1) return false;
	ZeroMemory(&bmpinfo, sizeof(bmpinfo));
	bmpinfo.bmiHeader.biSize = sizeof(BITMAPINFOHEADER);
	bmpinfo.bmiHeader.biWidth = w;
	bmpinfo.bmiHeader.biHeight = -h;
	bmpinfo.bmiHeader.biPlanes = 1;
	bmpinfo.bmiHeader.biBitCount = 32;
	bmpinfo.bmiHeader.biCompression = BI_RGB;
	bmphnd = CreateDIBSection(NULL, &bmpinfo, DIB_RGB_COLORS, (void **)(&adr), NULL, 0);
	if(!bmphnd) return false;
	hdc = GetDC(0);
	hdcnew = CreateCompatibleDC(hdc);
	ReleaseDC(0, hdc);
	if(!hdcnew){
		DeleteObject(bmphnd);
		return false;
	}
	if(m_PixelAdr){
		SelectObject(m_BmpHdc, m_BmpHndOld);
		DeleteObject(m_BmpHnd);
		DeleteDC(m_BmpHdc);
		m_PixelAdr = NULL;
	}
	m_BmpInfo = bmpinfo;
	m_PixelAdr = adr;
	m_BmpHnd = bmphnd;
	m_BmpHdc = hdcnew;
	m_BmpHndOld = (HBITMAP)SelectObject(m_BmpHdc, m_BmpHnd);
	m_Width = w;
	m_Height = h;
	ZeroMemory(m_PixelAdr, w*h*4);
	return true;
}

int CPixelbit::GetWidth(){
	return m_Width;
}

int CPixelbit::GetHeight(){
	return m_Height;
}

HDC CPixelbit::GetHDC(){
	return m_BmpHdc;
}

DWORD CPixelbit::GetPixel(int x, int y){
	return m_PixelAdr[x+y*m_Width];
}

DWORD CPixelbit::GetRGB(int x, int y){
	return 0x00ffffff&GetPixel(x, y);
}

DWORD CPixelbit::GetBGR(int x, int y){
	return (GetB(x, y)<<16)+(GetG(x, y)<<8)+GetR(x, y);;
}

BYTE CPixelbit::GetR(int x, int y){
	return (BYTE)((0x00ff0000&GetPixel(x, y))>>16);
}

BYTE CPixelbit::GetG(int x, int y){
	return (BYTE)((0x0000ff00&GetPixel(x, y))>>8);
}

BYTE CPixelbit::GetB(int x, int y){
	return (BYTE)(0x000000ff&GetPixel(x, y));
}

bool CPixelbit::IsColorKey(int x, int y){
	return GetRGB(x, y)==m_ColorKey;
}

bool CPixelbit::SetPixel(int x, int y, DWORD color){
	m_PixelAdr[x+y*m_Width] = color;
	return true;
}

bool CPixelbit::Set4BGR24(int x, int y, int w, int h, void *ptr){
	int i, j, len;
	BYTE *bgr = (BYTE *)(ptr);
	if(w*h<1 || !ptr || x+w<=0 || y+h<=0 || x>=w|| y>=h) return false;
	if(w % 4==0) len = w*3; else len = w*3+(4-(w*3) % 4);
	for(i = 0; i<h; i++){
		for(j = 0; j<w; j++){
			if(x+j>=0 && x+j<m_Width && y+i>=0 && y+i<m_Height){
				DWORD dwRGB = MAKEPIXEL(bgr[j*3+(h-i-1)*len+2], bgr[j*3+(h-i-1)*len+1], bgr[j*3+(h-i-1)*len]);
				SetRGB(x+j, y+i, dwRGB);
			}
		}
	}
	return true;
}

bool CPixelbit::Set4BGR32(int x, int y, int w, int h, void *ptr){
	int i, j, len;
	BYTE *bgr = (BYTE *)(ptr);
	if(w*h<1 || !ptr || x+w<=0 || y+h<=0 || x>=w|| y>=h) return false;
	len = w*4;
	for(i = 0; i<h; i++){
		for(j = 0; j<w; j++){
			if(x+j>=0 && x+j<m_Width && y+i>=0 && y+i<m_Height){
				DWORD dwRGB = MAKEPIXEL(bgr[j*4+(h-i-1)*len+2], bgr[j*4+(h-i-1)*len+1], bgr[j*4+(h-i-1)*len]);
				SetRGB(x+j, y+i, dwRGB);
			}
		}
	}
	return true;
}

bool CPixelbit::SetRGB(int x, int t, DWORD col){
	return SetPixel(x, t, 0x00ffffff&col);
}

bool CPixelbit::SetR(int x, int y, int v){
	if(v<0) v = 0;
	if(v>255) v = 255;
	SetPixel(x, y, (0xff00ffff&GetPixel(x, y))|(v<<16));
	return true;
}

bool CPixelbit::SetG(int x, int y, int v){
	if(v<0) v = 0;
	if(v>255) v = 255;
	SetPixel(x, y, (0xffff00ff&GetPixel(x, y))|(v<<8));
	return true;
}

bool CPixelbit::SetB(int x, int y, int v){
	if(v<0) v = 0;
	if(v>255) v = 255;
	SetPixel(x, y, (0xffffff00&GetPixel(x, y))|v);
	return true;
}

bool CPixelbit::AddR(int x, int y, int v){
	SetR(x, y, GetR(x, y)+v);
	return true;
}

bool CPixelbit::AddG(int x, int y, int v){
	SetG(x, y, GetG(x, y)+v);
	return true;
}

bool CPixelbit::AddB(int x, int y, int v){
	SetB(x, y, GetB(x, y)+v);
	return true;
}

DWORD CPixelbit::GetColorKey(){
	return m_ColorKey;
}

bool CPixelbit::SetColorKey(DWORD col){
	m_ColorKey = col;
	return true;
}

DWORD *CPixelbit::GetScanLine(int y){
	if(y<0 || y>=m_Height) return NULL;
	return m_PixelAdr+m_Width*y;
}

bool CPixelbit::FixLocation(int &dx, int &dy, int &sx, int &sy, int &w, int &h, int tw, int th){
	if(dx<=-m_Width || dy<=-m_Height || dx>=tw || dy>=th) return false;
	if(dx<0){
		w += dx;
		sx -= dx;
		dx = 0;
	}
	if(dy<0){
		h += dy;
		sy -= dy;
		dy = 0;
	}
	if(dx+w>tw) w = tw-dx;
	if(dy+h>th) h = th-dy;
	return w>0 && h>0;
}

bool CPixelbit::WindowStamp(HDC hdc, int dx, int dy, int sx, int sy, int w, int h){
	return SetDIBitsToDevice(hdc, dx, dy, w, h, sx, m_Height-sy-h, 0, m_Height, m_PixelAdr, &m_BmpInfo, DIB_RGB_COLORS) ? true : false;
}

bool CPixelbit::PlainStamp(CPixelbit *dest, int dx, int dy, int sx, int sy, int w, int h){
	int i, j, tx, ty;
	if(m_ColorKey==0xffffffff) return WindowStamp(dest->GetHDC(), dx, dy, sx, sy, w, h);
	if(!FixLocation(dx, dy, sx, sy, w, h, dest->GetWidth(), dest->GetHeight())) return false;
	for(i = 0; i<h; i++) for(j = 0; j<w; j++)
		if(!IsColorKey(tx = sx+j, ty = sy+i)) dest->SetPixel(dx+j, dy+i, GetPixel(tx, ty));
	return true;
}

bool CPixelbit::HalfStamp(CPixelbit *dest, int dx, int dy, int sx, int sy, int w, int h){
	int i, j, tx, ty;
	DWORD c1, c2, col;
	if(!FixLocation(dx, dy, sx, sy, w, h, dest->GetWidth(), dest->GetHeight())) return false;
	for(i = 0; i<h; i++){
		for(j = 0; j<w; j++){
			if(!IsColorKey(tx = sx+j, ty = sy+i)){
				c1 = GetPixel(tx, ty);
				c2 = dest->GetPixel(tx = dx+j, ty = dy+i);
				col = (((0x00fefefe&c1)+(0x00fefefe&c2))>>1)+(0x00010101&c1&c2);
				dest->SetPixel(tx, ty, col);
			}
		}
	}
	return true;
}

bool CPixelbit::FillRect(int x, int y, int w, int h, DWORD c){
	int i, j, tx = 0, ty = 0;
	if(!FixLocation(x, y, tx, ty, w, h, m_Width, m_Height)) return false;
	tx = x+w;
	ty = y+h;
	for(i = y; i<ty; i++) for(j = x; j<tx; j++) SetPixel(j, i, c);
	return true;
}

bool CPixelbit::Load(char *fname){
	BYTE *buf;
	HANDLE fh;
	DWORD load;
	if(!fname) return false;
	fh = CreateFile(fname, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	buf = new BYTE[GetFileSize(fh, NULL)];
	ReadFile(fh, buf, GetFileSize(fh, NULL), &load, NULL);
	CloseHandle(fh);
	if(Assign(buf)){
		delete [] buf;
		return true;
	}else{
		delete [] buf;
		return false;
	}
}

bool CPixelbit::Save(char *fname){
	int i, j, len;
	BYTE *buf, *pixel;
	LPBITMAPINFOHEADER head;
	HANDLE fh;
	DWORD wrote, size;
	if(!fname || m_Width*m_Height<1) return false;
	len = m_Width*3;
	if(len%4) len = m_Width*3+(4-(m_Width*3)%4);
	size = sizeof(BITMAPFILEHEADER)+sizeof(BITMAPINFOHEADER)+len*m_Height;
	buf = new BYTE[size];
	ZeroMemory(buf, size);
	buf[0] = 'B';
	buf[1] = 'M';
	CopyMemory(buf+2, &size, 4);
	buf[10] = 54;
	head = (LPBITMAPINFOHEADER)(buf+14);
	pixel = buf+54;
	head->biSize = 40;
	head->biWidth = m_Width;
	head->biHeight = m_Height;
	head->biPlanes = 1;
	head->biBitCount = 24;
	for(i = 0; i<m_Height; i++){
		for(j = 0;j <m_Width; j++){
			pixel[j*3+i*len] = GetB(j, m_Height-i-1);
			pixel[j*3+i*len+1] = GetG(j, m_Height-i-1);
			pixel[j*3+i*len+2] = GetR(j, m_Height-i-1);
		}
	}
	fh = CreateFile(fname, GENERIC_WRITE, 0, NULL, CREATE_ALWAYS, FILE_ATTRIBUTE_NORMAL, NULL);
	WriteFile(fh, buf, size, &wrote, NULL);
	CloseHandle(fh);
	delete [] buf;
	return wrote==size;
}

RECT MakeRect(int x, int y, int w, int h){
	RECT rect;
	rect.left = x;
	rect.top = y;
	rect.right = x+w;
	rect.bottom = y+h;
	return rect;
}
