//@@@@@@@@@@@@@@@@@@@@@@Copyright (c) 2002 Midikyou
#include"headers.h"
#include"debug.h"
#include"graphic.h"
#include"texture.h"
#include"mesh.h"

#include<rmxfguid.h>
#include<rmxftmpl.h>

/*============================================================================
  RXgN^
============================================================================*/
CXFile::CXFile()
{
	m_pXF = NULL;
	m_pPtr= NULL;
}
/*============================================================================
  fXgN^
============================================================================*/
CXFile::~CXFile()
{
	Close();
}
/*============================================================================
  t@C^\[X̃I[v
============================================================================*/
BOOL CXFile::Open(LPCSTR strSrc,BOOL fRes)
{
	HRESULT hr;

	DirectXFileCreate(&m_pXF);
	m_pXF->RegisterTemplates((LPVOID)D3DRM_XTEMPLATES, D3DRM_XTEMPLATE_BYTES);

	if(fRes)
	{
		DXFILELOADRESOURCE res;

		res.hModule = NULL;
		res.lpName  = strSrc;
		res.lpType  = "X";

		hr = m_pXF->CreateEnumObject(
			(LPVOID)&res, DXFILELOAD_FROMRESOURCE, &m_pPtr
		);
	}
	else
	{
		hr = m_pXF->CreateEnumObject(
			(LPVOID)strSrc,	DXFILELOAD_FROMFILE, &m_pPtr
		);
	}
	return hr == DXFILE_OK;
}
/*============================================================================
  ̃f[^擾
============================================================================*/
BOOL CXFile::GetNextData(LPDIRECTXFILEDATA* ppDat)
{
	HRESULT hr;

	hr = m_pPtr->GetNextDataObject(ppDat);

	return hr == DXFILE_OK;
}
/*============================================================================
  ŏKw̃bV擾
============================================================================*/
BOOL CXFile::GetTopMesh(LPDIRECTXFILEDATA* ppDat)
{
	const GUID *type;

	while(1)
	{
		if(!GetNextData(ppDat)) return FALSE;

		(*ppDat)->GetType(&type);

		if(*type == TID_D3DRMMesh) break;
		else RELEASE(*ppDat); //bVȊO͉
	}
	return TRUE;
}
/*============================================================================
  N[Y
============================================================================*/
void CXFile::Close()
{
	RELEASE(m_pPtr);
	RELEASE(m_pXF);
}

CTexList CMesh::s_texList; //eNX`Xg
/*============================================================================
  RXgN^
============================================================================*/
CMesh::CMesh()
{
	m_pMesh	= NULL;
	m_pMat	= NULL;
	m_pTex	= NULL;
}
/*============================================================================
  fXgN^
============================================================================*/
CMesh::~CMesh()
{
	Free();
}
/*============================================================================
  wt@C̓Ǎ

  strFile	:t@C
  cTrans	:eNX`[̓ߐF
  nMipLv	:~bv}bvku
============================================================================*/
BOOL CMesh::Load(char* strFile,D3DCOLOR cTrans,int nMipLv)
{
	//Ȃ
	if(m_pMesh) Free();

	Debug("load(%s) ... ",strFile);
	m_strName = strFile;

	//t@CǍ
    LPD3DXBUFFER pBuf;
    LPD3DXBUFFER pAdj;

	HRESULT hr;

	hr = D3DXLoadMeshFromX(
			strFile, D3DXMESH_SYSTEMMEM/*LocklƂꂪxXg*/, 
			sv3.pDev, &pAdj, &pBuf, &m_dwNumMat, &m_pMesh
		);

	if(FAILED(hr))
	{
		Debug("failed.\n");
		return FALSE;
	}
	Debug("ok.\n");

	//}eA̎擾AeNX`[̃[h
    D3DXMATERIAL* pMat = (D3DXMATERIAL*)pBuf->GetBufferPointer();

    m_pMat = new MAT8[m_dwNumMat];
    m_pTex = new LPTEX8[m_dwNumMat];

    for(DWORD i=0;i<m_dwNumMat;i++ )
    {
        m_pMat[i] = pMat[i].MatD3D;
        m_pMat[i].Ambient = m_pMat[i].Diffuse;
		m_pTex[i] = NULL;

		if(pMat[i].pTextureFilename == NULL) continue;

		//XgɂΎQƁAȂ΃[hĒǉ
		m_pTex[i] = s_texList.Get(FALSE, pMat[i].pTextureFilename, cTrans, nMipLv);
    }
    RELEASE(pBuf);

	//ĚvZ
	ComputeBoundary();

	//bV̍œK
	LPD3DXMESH pMeshOpt;
	
	hr = m_pMesh->Optimize(
	  D3DXMESHOPT_ATTRSORT | D3DXMESHOPT_COMPACT | D3DXMESHOPT_VERTEXCACHE,
	  (DWORD*)pAdj->GetBufferPointer(),
	  NULL, NULL, NULL, &pMeshOpt
	);
	if( SUCCEEDED(hr) ){
		m_pMesh->Release();
		m_pMesh = pMeshOpt;
	}
	else Debug("optimization failed.\n");
	
	RELEASE(pAdj);
	return TRUE;
}
//\[X
BOOL CMesh::LoadResource(char* strRes,D3DCOLOR cTrans,int nMipLv)
{
	//Ȃ
	if(m_pMesh) Free();

	Debug("load(%s) ... ",strRes);
	m_strName = strRes;

	CXFile xfile;

	if(!xfile.Open(strRes,TRUE))
	{
		Debug("open failed.\n");
		return FALSE;
	}

	LPDIRECTXFILEDATA pDat;

	if(!xfile.GetTopMesh(&pDat))
	{
		Debug("mesh is not found.\n");
		return FALSE;
	}
	HRESULT hr;

    LPD3DXBUFFER pBuf;
    LPD3DXBUFFER pAdj;

	hr = D3DXLoadMeshFromXof(
			pDat, D3DXMESH_SYSTEMMEM, 
			sv3.pDev, &pAdj, &pBuf, &m_dwNumMat, &m_pMesh
		);
	RELEASE(pDat);
	xfile.Close();

	if(FAILED(hr))
	{
		Debug("failed.\n");
		return FALSE;
	}
	Debug("ok.\n");

	//}eA̎擾AeNX`[̃[h
    D3DXMATERIAL* pMat = (D3DXMATERIAL*)pBuf->GetBufferPointer();

    m_pMat = new MAT8[m_dwNumMat];
    m_pTex = new LPTEX8[m_dwNumMat];

    for(DWORD i=0;i<m_dwNumMat;i++ )
    {
        m_pMat[i] = pMat[i].MatD3D;
        m_pMat[i].Ambient = m_pMat[i].Diffuse;
		m_pTex[i] = NULL;

		if(pMat[i].pTextureFilename == NULL) continue;

		m_pTex[i] = s_texList.Get(TRUE, pMat[i].pTextureFilename, cTrans, nMipLv);
    }
    RELEASE(pBuf);

	//ĚvZ
	ComputeBoundary();

	//bV̍œK
	LPD3DXMESH pMeshOpt;

	hr = m_pMesh->Optimize(
	  D3DXMESHOPT_ATTRSORT | D3DXMESHOPT_COMPACT | D3DXMESHOPT_VERTEXCACHE,
	  (DWORD*)pAdj->GetBufferPointer(),
	  NULL, NULL, NULL, &pMeshOpt
	);
	if( SUCCEEDED(hr) ){
		m_pMesh->Release();
		m_pMesh = pMeshOpt;
	}
	else Debug("optimization failed.\n");

	RELEASE(pAdj);
	return TRUE;
}
/*============================================================================
  ̍쐬

  sl	:XCX
  st	:X^bN
  cv	:F
============================================================================*/
BOOL CMesh::CreateSphere(float r,UINT sl,UINT st,D3DCOLORVALUE cv)
{
	//Ȃ
	if(m_pMesh) Free();

	if(FAILED( D3DXCreateSphere(sv3.pDev, r, sl, st, &m_pMesh, NULL) ))
	{
		Debug("D3DXCreateSphere\n");
		return FALSE;
	}
	m_strName = "";
    m_pMat = new MAT8[1];
    m_pTex = new LPTEX8[1];

 	m_dwNumMat = 1;
	m_pMat[0].Diffuse	= m_pMat[0].Ambient = cv;
	m_pMat[0].Specular	= m_pMat[0].Emissive = MAKE_CV(0,0,0,0);
	m_pMat[0].Power		= 0.0f;
	m_pTex[0]			= NULL;
	
	ComputeBoundary();
	return TRUE;
}
/*============================================================================
  ̂̍쐬

  x		:XTCY
  y		:YTCY
  z		:ZTCY
  cv	:F
============================================================================*/
BOOL CMesh::CreateBox(float x,float y,float z,D3DCOLORVALUE cv)
{
	//Ȃ
	if(m_pMesh) Free();

	if(FAILED( D3DXCreateBox(sv3.pDev, x, y, z, &m_pMesh, NULL) ))
	{
		Debug("D3DXCreateSphere\n");
		return FALSE;
	}
 	m_strName = "";
	m_dwNumMat = 1;
	m_pMat = new MAT8[1];
    m_pTex = new LPTEX8[1];

	m_pMat[0].Diffuse	= m_pMat[0].Ambient = cv;
	m_pMat[0].Specular	= m_pMat[0].Emissive = MAKE_CV(0,0,0,0);
	m_pMat[0].Power		= 0.0f;
	m_pTex[0]			= NULL;
	
	ComputeBoundary();
	return TRUE;
}
/*============================================================================
  ̂̍쐬

  x		:XTCY
  y		:YTCY
  z		:ZTCY
  cv	:F
============================================================================*/
BOOL CMesh::CreateTeapot(D3DCOLORVALUE cv)
{
	//Ȃ
	if(m_pMesh) Free();

	if(FAILED( D3DXCreateTeapot(sv3.pDev, &m_pMesh, NULL) ))
	{
		Debug("D3DXCreateTeapot\n");
		return FALSE;
	}
	m_strName = "";
 	m_dwNumMat = 1;
	m_pMat = new MAT8[1];
    m_pTex = new LPTEX8[1];

	m_pMat[0].Diffuse  = m_pMat[0].Ambient  = cv;
	m_pMat[0].Specular = m_pMat[0].Emissive = MAKE_CV(0,0,0,0);
	m_pMat[0].Power    = 0.0f;
	m_pTex[0]          = NULL;
	
	ComputeBoundary();
	return TRUE;
}
/*============================================================================
  bV̉
============================================================================*/
void CMesh::Free()
{
	if(m_strName != "")
	{
		Debug("release(%s)\n",m_strName.c_str());
		m_strName = "";
	}

	if(m_pMat)
	{
		delete[] m_pMat;
		m_pMat = NULL;
	}

	if(m_pTex)
	{
		//eNX`[̉
		for(DWORD i = 0;i < m_dwNumMat;i++)
		{
			s_texList.Release(m_pTex[i]);
		}
		//eNX`[Xg̍폜
		delete[] m_pTex;
		m_pTex = NULL;
	}
	RELEASE(m_pMesh);
}
/*============================================================================
  E{bNX^̌vZ
============================================================================*/
void CMesh::ComputeBoundary()
{
	LPBYTE	pVertex;
	DWORD	nVertex,fvf;

	m_pMesh->LockVertexBuffer(D3DLOCK_READONLY | D3DLOCK_NOSYSLOCK,&pVertex);
	nVertex	= m_pMesh->GetNumVertices();
	fvf		= m_pMesh->GetFVF();

	D3DXComputeBoundingBox(
		(VOID*)pVertex,
		nVertex,
		fvf,
		&m_min,
		&m_max
	);
	D3DXComputeBoundingSphere(
		(VOID*)pVertex,
		nVertex,
		fvf,
		&m_center,
		&m_radius
	);
	m_pMesh->UnlockVertexBuffer();
}
/*============================================================================
  _O

  pMtx	:Wϊs

@XVfoCXp[^	:[h}gNXA}eAAeNX`
============================================================================*/
void CMesh::Render(MTX4* pMtx)
{
	if(m_pMesh == NULL) return;

	devTransform(pMtx);

	for(DWORD i=0;i<m_dwNumMat;i++)
	{
		sv3.pDev->SetMaterial( &m_pMat[i] );
		sv3.pDev->SetTexture( 0, m_pTex[i] );
		m_pMesh->DrawSubset( i );
	}
}
/*============================================================================
  _O(eNX`w)

  pMtx	:Wϊs
  pTex	:eNX`

@XVfoCXp[^	:[h}gNXA}eAAeNX`
============================================================================*/
void CMesh::RenderT(MTX4* pMtx,LPTEX8 pTex)
{
	if(m_pMesh == NULL) return;

	devTransform(pMtx);

	for(DWORD i=0;i<m_dwNumMat;i++)
	{
		sv3.pDev->SetMaterial( &m_pMat[i] );
		sv3.pDev->SetTexture( 0, pTex );
		m_pMesh->DrawSubset( i );
	}
}
/*============================================================================
  _O(At@lw)

  pMtx	:Wϊs
  alpha :At@l

@XVfoCXp[^	:[h}gNXA}eAAeNX`
============================================================================*/
void CMesh::RenderA(MTX4* pMtx,float alpha)
{
	if(m_pMesh == NULL) return;

	devTransform(pMtx);

	D3DCOLORVALUE c1,c2;

	for(DWORD i=0;i<m_dwNumMat;i++)
	{
		c1		= m_pMat[i].Ambient;
		c2		= c1;
		c2.a	= alpha;
		m_pMat[i].Diffuse = m_pMat[i].Ambient = c2;

		sv3.pDev->SetMaterial( &m_pMat[i] );
		sv3.pDev->SetTexture( 0, m_pTex[i] );
		m_pMesh->DrawSubset( i );

		m_pMat[i].Diffuse = m_pMat[i].Ambient = c1;
	}
}
/*============================================================================
  _OiZj

  pMtx	:Wϊs
@aplus	:Zl

@XVfoCXp[^	:[h}gNXA}eAAeNX`
============================================================================*/
void CMesh::RenderAP(MTX4* pMtx,float aplus)
{
	if(m_pMesh == NULL) return;

	devTransform(pMtx);

	D3DCOLORVALUE c1,c2;

	for(DWORD i=0;i<m_dwNumMat;i++)
	{
		//lύX
		c1		= m_pMat[0].Ambient; // = m_pMat[0].Diffuse
		c2		= c1;
		c2.a	= max(-1.0f, min(1.0f, c2.a+aplus));
		m_pMat[i].Diffuse = m_pMat[i].Ambient = c2;

		sv3.pDev->SetMaterial( &m_pMat[i] );
		sv3.pDev->SetTexture( 0, m_pTex[i] );
		m_pMesh->DrawSubset( i );

		//l𕜌
		m_pMat[i].Diffuse = m_pMat[i].Ambient = c1;
	}
}
/*============================================================================
  _Oi}eAwAeȂǂŎgpj

  pMtx	:Wϊs
@mat	:}eA

@XVfoCXp[^	:[h}gNXA}eAAeNX`
============================================================================*/
void CMesh::RenderSC(MTX4* pMtx,D3DMATERIAL8* pMat)
{
	if(m_pMesh == NULL) return;

	devTransform(pMtx);

	for(DWORD i=0;i<m_dwNumMat;i++)
	{
		sv3.pDev->SetMaterial( pMat );
		sv3.pDev->SetTexture( 0, NULL );
		m_pMesh->DrawSubset( i );
	}
}
